@@ -3,7 +3,7 @@
 MyOpenGLView.m
 
 Oolite
-Copyright (C) 2004-2011 Giles C Williams and contributors
+Copyright (C) 2004-2013 Giles C Williams and contributors
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -45,6 +45,11 @@ static NSString * kOOLogKeyUp				= @"input.keyMapping.keyPress.keyUp";
 static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 
 
+static void GetDesiredCursorState(OOMouseInteractionMode mode, BOOL *outHidden, BOOL *outObscured);
+static void ApplyCursorState(OOMouseInteractionMode mode);
+static void UnapplyCursorState(OOMouseInteractionMode mode);
+
+
 @interface MyOpenGLView(Internal)
 
 - (int) translateKeyCode:(int)input;
@@ -54,10 +59,18 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 @end
 
 
-#if !OOLITE_SNOW_LEOPARD
-@interface NSResponder (SnowLeopard)
+#if !OOLITE_MAC_OS_X_10_7
+@interface NSView (Lion)
 
-- (void) setAcceptsTouchEvents:(BOOL)value;
+- (BOOL) wantsBestResolutionOpenGLSurface;
+- (void) setWantsBestResolutionOpenGLSurface:(BOOL)flag;
+
+- (NSPoint) convertPointToBacking:(NSPoint)aPoint;
+- (NSPoint) convertPointFromBacking:(NSPoint)aPoint;
+- (NSSize) convertSizeToBacking:(NSSize)aSize;
+- (NSSize) convertSizeFromBacking:(NSSize)aSize;
+- (NSRect) convertRectToBacking:(NSRect)aRect;
+- (NSRect) convertRectFromBacking:(NSRect)aRect;
 
 @end
 #endif
@@ -74,13 +87,11 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 	}
 #endif
 	
-	if (!(self = [super initWithFrame:frameRect]))  return nil;
-	
 	// Pixel Format Attributes for the View-based (non-FullScreen) NSOpenGLContext
 	NSOpenGLPixelFormatAttribute attrs[] =
 	{
 		// Specify that we want a windowed OpenGL context.
-		// Must be first or we'll hit an assert in -[GameController goFullscreen:].
+		// Must be first or we'll hit an assert in the legacy fullscreen controller.
 		NSOpenGLPFAWindow,
 		
 		// We may be on a multi-display system (and each screen may be driven by a different renderer), so we need to specify which screen we want to take over.
@@ -108,26 +119,34 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 		0
 	};
 	
-	_pixelFormatAttributes = [[NSData alloc] initWithBytes:attrs length:sizeof attrs];
-	
 	// Create our non-FullScreen pixel format.
-	NSOpenGLPixelFormat* pixelFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attrs] autorelease];
+	NSOpenGLPixelFormat *pixelFormat = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attrs] autorelease];
 	
-	self = [super initWithFrame:frameRect pixelFormat:pixelFormat];
-	
-	virtualJoystickPosition = NSMakePoint(0.0,0.0);
-	if ([self respondsToSelector:@selector(setAcceptsTouchEvents:)])
+	if ((self = [super initWithFrame:frameRect pixelFormat:pixelFormat]))
 	{
-		[self setAcceptsTouchEvents:YES];
-	}
-	
-	typedString = [[NSMutableString alloc] initWithString:@""];
-	allowingStringInput = gvStringInputNo;
-	isAlphabetKeyDown = NO;
+		if ([self respondsToSelector:@selector(setAcceptsTouchEvents:)])
+		{
+			[self setAcceptsTouchEvents:YES];
+		}
 		
-	timeIntervalAtLastClick = [NSDate timeIntervalSinceReferenceDate];
-	
-	_virtualScreen = [[self openGLContext] currentVirtualScreen];
+		if ([self respondsToSelector:@selector(setWantsBestResolutionOpenGLSurface:)])
+		{
+			// Enable high resolution on Retina displays.
+			[self setWantsBestResolutionOpenGLSurface:YES];
+		}
+		
+		matrixManager = [[OOOpenGLMatrixManager alloc] init];
+		_pixelFormatAttributes = [[NSData alloc] initWithBytes:attrs length:sizeof attrs];
+		virtualJoystickPosition = NSMakePoint(0.0,0.0);
+		
+		typedString = [[NSMutableString alloc] initWithString:@""];
+		allowingStringInput = gvStringInputNo;
+		isAlphabetKeyDown = NO;
+			
+		timeIntervalAtLastClick = [NSDate timeIntervalSinceReferenceDate];
+		
+		_virtualScreen = [[self openGLContext] currentVirtualScreen];
+	}
 	
 	return self;
 }
@@ -137,26 +156,31 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 {
 	DESTROY(typedString);
 	DESTROY(_pixelFormatAttributes);
+	DESTROY(matrixManager);
 	
 	[super dealloc];
 }
 
 
-- (void) setStringInput: (enum StringInput) value
+- (void) setStringInput:(enum StringInput)value
 {
 	allowingStringInput = value;
 }
 
 
-- (void) allowStringInput: (BOOL) value
+- (void) allowStringInput:(BOOL)value
 {
 	if (value)
+	{
 		allowingStringInput = gvStringInputAlpha;
+	}
 	else
+	{
 		allowingStringInput = gvStringInputNo;
+	}
 }
 
--(enum StringInput) allowingStringInput
+- (enum StringInput) allowingStringInput
 {
 	return allowingStringInput;
 }
@@ -174,7 +198,7 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
-- (void) setTypedString:(NSString*) value
+- (void) setTypedString:(NSString *)value
 {
 	[typedString setString:value];
 }
@@ -186,6 +210,12 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
+- (NSSize) backingViewSize
+{
+	return backingViewSize;
+}
+
+
 - (GLfloat) display_z
 {
 	return display_z;
@@ -204,7 +234,7 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
-- (GameController *)gameController
+- (GameController *) gameController
 {
 	return gameController;
 }
@@ -216,12 +246,27 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
-- (void) updateScreen
+- (void) noteMouseInteractionModeChangedFrom:(OOMouseInteractionMode)oldMode to:(OOMouseInteractionMode)newMode
 {
-	[self drawRect:NSMakeRect(0, 0, viewSize.width, viewSize.height)];
+	UnapplyCursorState(oldMode);
+	ApplyCursorState(newMode);
 }
 
 
+- (void) updateScreen
+{
+	if ([[self window] isVisible])
+	{
+		[self drawRect:NSMakeRect(0, 0, viewSize.width, viewSize.height)];
+	}
+}
+
+- (void) pollControls
+{
+	if ([NSDate timeIntervalSinceReferenceDate] > timeIntervalAtLastMouseWheel + OOMOUSEWHEEL_EVENTS_DELAY_INTERVAL)
+		_mouseWheelDelta = 0.0f;
+}
+
 - (void) drawRect:(NSRect)rect
 {
 	if ((viewSize.width != [self frame].size.width)||(viewSize.height != [self frame].size.height)) // resized
@@ -287,7 +332,7 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
-- (void) initialiseGLWithSize:(NSSize) v_size
+- (void) initialiseGLWithSize:(NSSize)v_size
 {
 	viewSize = v_size;
 	if (viewSize.width/viewSize.height > 4.0/3.0) {
@@ -300,8 +345,18 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 		y_offset = 320.0 * viewSize.height/viewSize.width;
 	}
 	
-	[self openGLContext];
-	[[self gameController] setUpBasicOpenGLStateWithSize:viewSize];
+	if ([self respondsToSelector:@selector(convertSizeToBacking:)])
+	{
+		// High resolution mode support.
+		backingViewSize = [self convertSizeToBacking:viewSize];
+	}
+	else
+	{
+		backingViewSize = viewSize;
+	}
+	
+	[self openGLContext];	// Force lazy setup if needed.
+	[[self gameController] setUpBasicOpenGLStateWithSize:v_size];
 	[[self openGLContext] flushBuffer];
 	
 	m_glContextInitialized = YES;
@@ -314,11 +369,41 @@ static NSString * kOOLogKeyDown				= @"input.keyMapping.keyPress.keyDown";
 }
 
 
+#ifdef MAC_OS_X_VERSION_10_7	// If building against 10.7 SDK, where relevant symbols are defined...
+- (void) viewDidMoveToWindow
+{
+	/*	Subscribe to NSWindowDidChangeBackingPropertiesNotification on systems
+		which support it (10.7 and later). This notification fires when the
+		scale factor or colour space of the window's backing store changes.
+		We use it to track scale factor changes.
+	*/
+	if (&NSWindowDidChangeBackingPropertiesNotification != NULL && [self.window respondsToSelector:@selector(backingScaleFactor)])
+	{
+		[[NSNotificationCenter defaultCenter] addObserver:self
+												 selector:@selector(backingPropertiesChanged:)
+													 name:NSWindowDidChangeBackingPropertiesNotification
+												   object:self.window];
+		
+		// Also, ensure the initial state makes sense.
+		[self backingPropertiesChanged:nil];
+	}
+}
+
+
+- (void) backingPropertiesChanged:(NSNotification *)notification
+{
+	GLSetDisplayScaleFactor(self.window.backingScaleFactor);
+}
+#endif
+
+
 - (BOOL) snapShot:(NSString *)filename
 {
 	BOOL snapShotOK = YES;
-	int w = viewSize.width;
-	int h = viewSize.height;
+	NSSize v_size = backingViewSize;
+	
+	int w = v_size.width;
+	int h = v_size.height;
 	
 	if (w & 3)
 		w = w + 4 - (w & 3);
@@ -428,6 +513,14 @@ FAIL:
 }
 
 
+- (void) stringToClipboard:(NSString *)stringToCopy
+{
+	NSPasteboard *clipBoard = [NSPasteboard generalPasteboard];
+	[clipBoard declareTypes:[NSArray arrayWithObject:NSPasteboardTypeString] owner:nil];
+	[clipBoard setString:stringToCopy forType:NSPasteboardTypeString];
+}
+
+
 - (BOOL) acceptsFirstResponder
 {
 	return YES;
@@ -548,16 +641,7 @@ FAIL:
 		if (!f12)
 		{
 			f12 = YES;
-			
-			if (![gameController inFullScreenMode])
-			{
-				// Command-F is handled by menu action in non-fullscreen mode; simulate.
-				[gameController performSelector:@selector(goFullscreen:) withObject:nil afterDelay:0.0];
-			}
-			else
-			{
-				commandF = YES;
-			}
+			[gameController performSelector:@selector(toggleFullScreenAction:) withObject:nil afterDelay:0.0];
 		}
 		
 		return;
@@ -588,14 +672,20 @@ FAIL:
 				}
 			}
 			
-			// full input for load-save screen
-			if (allowingStringInput == gvStringInputAll)
+			// full input for load-save screen or 'all' input
+			if (allowingStringInput >= gvStringInputLoadSave)
 			{
+				// except '/' for loadsave
 				if (isprint(key) && key != '/')
 				{
 					isAlphabetKeyDown = YES;
 					[typedString appendFormat:@"%c", key];
 				}
+				else if (key == '/' && allowingStringInput == gvStringInputAll)
+				{
+					isAlphabetKeyDown = YES;
+					[typedString appendFormat:@"%c", key];
+				}
 			}
 			
 		}
@@ -612,11 +702,15 @@ FAIL:
 /* Capture shift, ctrl, opt and command press & release */
 - (void)flagsChanged:(NSEvent *)theEvent
 {
-	int flags = [theEvent modifierFlags];
+	NSUInteger flags = [theEvent modifierFlags];
 	opt = (flags & NSAlternateKeyMask) ? YES : NO;
 	ctrl = (flags & NSControlKeyMask) ? YES : NO;
 	command = (flags & NSCommandKeyMask) ? YES : NO;
 	shift = ( flags & NSShiftKeyMask ) ? YES : NO;
+	if ([theEvent keyCode] == 0x39) // 57 = key code for caps lock
+	{
+		capsLockOn = (flags & NSAlphaShiftKeyMask) ? YES : NO;
+	}
 }
 
 
@@ -646,11 +740,19 @@ FAIL:
 }
 
 
+- (void)resetMouse
+{
+	CGPoint centerPoint = CGPointMake(viewSize.width / 2.0, viewSize.height / 2.0);
+	CGWarpMouseCursorPosition(centerPoint);
+	[self setVirtualJoystick:0.0 :0.0];
+}
+
+
 - (void)mouseMoved:(NSEvent *)theEvent
 {
 	double mx = [theEvent locationInWindow].x - viewSize.width/2.0;
 	double my = [theEvent locationInWindow].y - viewSize.height/2.0;
-		
+
 	if (display_z > 640.0)
 	{
 		mx /= viewSize.width * MAIN_GUI_PIXEL_WIDTH / display_z;
@@ -671,6 +773,29 @@ FAIL:
 	[self mouseMoved:theEvent];
 }
 
+- (void)scrollWheel:(NSEvent *)theEvent
+{
+	float dy = [theEvent scrollingDeltaY];
+	
+	if (dy == 0)
+		return;
+	
+	if (dy > 0)
+	{
+		if (_mouseWheelDelta >= 0)
+			_mouseWheelDelta += dy;
+		else
+			_mouseWheelDelta = 0;
+	}
+	else
+	{
+		if (_mouseWheelDelta <= 0)
+			_mouseWheelDelta += dy;
+		else
+			_mouseWheelDelta = 0;
+	}
+	timeIntervalAtLastMouseWheel = [NSDate timeIntervalSinceReferenceDate];
+}
 
 - (void) otherMouseDragged:(NSEvent *)theEvent
 {
@@ -680,7 +805,7 @@ FAIL:
 
 - (void) rightMouseDown:(NSEvent *)theEvent
 {
-	[self recenterVirtualJoystick];
+	[self resetMouse];
 }
 
 
@@ -902,12 +1027,41 @@ FAIL:
 }
 
 
+- (BOOL) isCapsLockOn
+{
+	return capsLockOn;
+}
+
+
 - (int) numKeys
 {
 	return NUM_KEYS;
 }
 
 
+- (int) mouseWheelState
+{
+	if (_mouseWheelDelta > 0.0f)
+		return gvMouseWheelUp;
+	else if (_mouseWheelDelta < 0.0f)
+		return gvMouseWheelDown;
+	else
+		return gvMouseWheelNeutral;
+}
+
+
+- (float) mouseWheelDelta
+{
+	return _mouseWheelDelta / OOMOUSEWHEEL_DELTA;
+}
+
+
+- (void) setMouseWheelDelta: (float) newWheelDelta
+{
+	_mouseWheelDelta = newWheelDelta * OOMOUSEWHEEL_DELTA;
+}
+
+
 - (BOOL) isCommandQDown
 {
 	return commandQ;
@@ -936,14 +1090,18 @@ FAIL:
 	return KEYMAP_GET(map, 56) || KEYMAP_GET(map, 60);	// Left shift or right shift -- although 60 shouldn't occur.
 }
 
+- (OOOpenGLMatrixManager *) getOpenGLMatrixManager
+{
+	return matrixManager;
+}
 
 #ifndef NDEBUG
 // General image-dumping method.
 - (void) dumpRGBAToFileNamed:(NSString *)name
 					   bytes:(uint8_t *)bytes
-					   width:(OOUInteger)width
-					  height:(OOUInteger)height
-					rowBytes:(OOUInteger)rowBytes
+					   width:(NSUInteger)width
+					  height:(NSUInteger)height
+					rowBytes:(NSUInteger)rowBytes
 {
 	if (name == nil || bytes == NULL || width == 0 || height == 0 || rowBytes < width * 4)  return;
 	
@@ -963,15 +1121,15 @@ FAIL:
 		[bitmap autorelease];
 		
 		NSString *filepath = [[[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent] stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"png"];
-		[[bitmap representationUsingType:NSPNGFileType properties:nil] writeToFile:filepath atomically:YES];
+		[[bitmap representationUsingType:NSPNGFileType properties:[NSDictionary dictionary]] writeToFile:filepath atomically:YES];
 	}
 }
 
 - (void) dumpRGBToFileNamed:(NSString *)name
 					  bytes:(uint8_t *)bytes
-					  width:(OOUInteger)width
-					 height:(OOUInteger)height
-				   rowBytes:(OOUInteger)rowBytes
+					  width:(NSUInteger)width
+					 height:(NSUInteger)height
+				   rowBytes:(NSUInteger)rowBytes
 {
 	if (name == nil || bytes == NULL || width == 0 || height == 0 || rowBytes < width * 3)  return;
 	
@@ -991,16 +1149,16 @@ FAIL:
 		[bitmap autorelease];
 		
 		NSString *filepath = [[[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent] stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"png"];
-		[[bitmap representationUsingType:NSPNGFileType properties:nil] writeToFile:filepath atomically:YES];
+		[[bitmap representationUsingType:NSPNGFileType properties:[NSDictionary dictionary]] writeToFile:filepath atomically:YES];
 	}
 }
 
 
 - (void) dumpGrayToFileNamed:(NSString *)name
 					   bytes:(uint8_t *)bytes
-					   width:(OOUInteger)width
-					  height:(OOUInteger)height
-					rowBytes:(OOUInteger)rowBytes
+					   width:(NSUInteger)width
+					  height:(NSUInteger)height
+					rowBytes:(NSUInteger)rowBytes
 {
 	if (name == nil || bytes == NULL || width == 0 || height == 0 || rowBytes < width)  return;
 	
@@ -1020,16 +1178,16 @@ FAIL:
 		[bitmap autorelease];
 		
 		NSString *filepath = [[[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent] stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"png"];
-		[[bitmap representationUsingType:NSPNGFileType properties:nil] writeToFile:filepath atomically:YES];
+		[[bitmap representationUsingType:NSPNGFileType properties:[NSDictionary dictionary]] writeToFile:filepath atomically:YES];
 	}
 }
 
 
 - (void) dumpGrayAlphaToFileNamed:(NSString *)name
 							bytes:(uint8_t *)bytes
-							width:(OOUInteger)width
-						   height:(OOUInteger)height
-						 rowBytes:(OOUInteger)rowBytes
+							width:(NSUInteger)width
+						   height:(NSUInteger)height
+						 rowBytes:(NSUInteger)rowBytes
 {
 	if (name == nil || bytes == NULL || width == 0 || height == 0 || rowBytes < width * 2)  return;
 	
@@ -1049,7 +1207,7 @@ FAIL:
 		[bitmap autorelease];
 		
 		NSString *filepath = [[[[[NSBundle mainBundle] bundlePath] stringByDeletingLastPathComponent] stringByAppendingPathComponent:name] stringByAppendingPathExtension:@"png"];
-		[[bitmap representationUsingType:NSPNGFileType properties:nil] writeToFile:filepath atomically:YES];
+		[[bitmap representationUsingType:NSPNGFileType properties:[NSDictionary dictionary]] writeToFile:filepath atomically:YES];
 	}
 }
 
@@ -1057,14 +1215,14 @@ FAIL:
 - (void) dumpRGBAToRGBFileNamed:(NSString *)rgbName
 			   andGrayFileNamed:(NSString *)grayName
 						  bytes:(uint8_t *)bytes
-						  width:(OOUInteger)width
-						 height:(OOUInteger)height
-					   rowBytes:(OOUInteger)rowBytes
+						  width:(NSUInteger)width
+						 height:(NSUInteger)height
+					   rowBytes:(NSUInteger)rowBytes
 {
 	if ((rgbName == nil && grayName == nil) || bytes == NULL || width == 0 || height == 0 || rowBytes < width * 4)  return;
 	
 	uint8_t				*rgbBytes, *rgbPx, *grayBytes, *grayPx, *srcPx;
-	OOUInteger			x, y;
+	NSUInteger			x, y;
 	BOOL				trivalAlpha = YES;
 	
 	rgbPx = rgbBytes = malloc(width * height * 3);
@@ -1111,4 +1269,54 @@ FAIL:
 
 #endif
 
+
+static void GetDesiredCursorState(OOMouseInteractionMode mode, BOOL *outHidden, BOOL *outObscured)
+{
+	NSCParameterAssert(outHidden != NULL && outObscured != NULL);
+	
+	*outHidden = (mode == MOUSE_MODE_FLIGHT_WITH_MOUSE_CONTROL);
+	*outObscured = (mode == MOUSE_MODE_FLIGHT_NO_MOUSE_CONTROL);
+}
+
+
+static void ApplyCursorState(OOMouseInteractionMode mode)
+{
+	BOOL hidden, obscured;
+	GetDesiredCursorState(mode, &hidden, &obscured);
+	if (hidden)  [NSCursor hide];
+	if (obscured)  [NSCursor setHiddenUntilMouseMoves:YES];
+}
+
+
+static void UnapplyCursorState(OOMouseInteractionMode mode)
+{
+	BOOL hidden, obscured;
+	GetDesiredCursorState(mode, &hidden, &obscured);
+	if (hidden)  [NSCursor unhide];
+	if (obscured)  [NSCursor setHiddenUntilMouseMoves:NO];
+}
+
+
+- (void) setGammaValue: (float) value
+{
+	// no-op
+}
+
+- (float) gammaValue
+{
+	return 1.0;
+}
+
+
+- (void) setFov:(float)value fromFraction:(BOOL)fromFraction
+{
+	_fov = fromFraction ? value : tanf((value / 2.0f) * M_PI / 180.0f);
+}
+
+
+- (float) fov:(BOOL)inFraction
+{
+	return inFraction ? _fov : 2 * atanf(_fov) * 180.0f / M_PI;
+}
+
 @end
