@@ -16,7 +16,7 @@ present.
 
 
 Oolite
-Copyright (C) 2004-2011 Giles C Williams and contributors
+Copyright (C) 2004-2013 Giles C Williams and contributors
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -37,7 +37,7 @@ MA 02110-1301, USA.
 
 #import "OOMesh.h"
 #import "Universe.h"
-#import "Geometry.h"
+#import "OOMeshToOctreeConverter.h"
 #import "ResourceManager.h"
 #import "Entity.h"		// for NO_DRAW_DISTANCE_FACTOR.
 #import "Octree.h"
@@ -54,7 +54,7 @@ MA 02110-1301, USA.
 #import "NSObjectOOExtensions.h"
 
 #import "OOJavaScriptEngine.h"
-
+#import "OODebugStandards.h"
 
 // If set, collision octree depth varies depending on the size of the mesh.
 #define ADAPTIVE_OCTREE_DEPTH		1
@@ -66,12 +66,10 @@ MA 02110-1301, USA.
 enum
 {
 	kBaseOctreeDepth				= 5,	// 32x32x32
-	kMaxOctreeDepth					= 7,	// 128x128x128
+//	kMaxOctreeDepth declared in Octree.h.
 	kSmallOctreeDepth				= 4,	// 16x16x16
-	kVerySmallOctreeDepth			= 3,	// 8x8x8
 	kOctreeSizeThreshold			= 900,	// Size at which we start increasing octree depth
-	kOctreeSmallSizeThreshold		= 50,
-	kOctreeVerySmallSizeThreshold	= 15
+	kOctreeSmallSizeThreshold		= 20
 };
 
 
@@ -84,8 +82,6 @@ typedef enum
 
 
 static NSString * const kOOLogMeshDataNotFound				= @"mesh.load.failed.fileNotFound";
-static NSString * const kOOLogMeshTooManyVertices			= @"mesh.load.failed.tooManyVertices";
-static NSString * const kOOLogMeshTooManyFaces				= @"mesh.load.failed.tooManyFaces";
 static NSString * const kOOLogMeshTooManyMaterials			= @"mesh.load.failed.tooManyMaterials";
 
 
@@ -126,9 +122,9 @@ typedef struct VertexFaceRef
 } VertexFaceRef;
 
 
-static void VFRAddFace(VertexFaceRef *vfr, OOUInteger index);
-static OOUInteger VFRGetCount(VertexFaceRef *vfr);
-static OOUInteger VFRGetFaceAtIndex(VertexFaceRef *vfr, OOUInteger index);
+static void VFRAddFace(VertexFaceRef *vfr, NSUInteger index);
+static NSUInteger VFRGetCount(VertexFaceRef *vfr);
+static NSUInteger VFRGetFaceAtIndex(VertexFaceRef *vfr, NSUInteger index);
 
 
 @interface OOMesh (Private) <NSMutableCopying, OOGraphicsResetClient>
@@ -139,9 +135,11 @@ materialDictionary:(NSDictionary *)materialDict
  shadersDictionary:(NSDictionary *)shadersDict
 			smooth:(BOOL)smooth
 	  shaderMacros:(NSDictionary *)macros
-shaderBindingTarget:(id<OOWeakReferenceSupport>)object;
+shaderBindingTarget:(id<OOWeakReferenceSupport>)object
+	   scaleFactor:(float)scale
+	cacheWriteable:(BOOL)cacheWriteable;
 
-- (BOOL) loadData:(NSString *)filename;
+- (BOOL) loadData:(NSString *)filename scaleFactor:(float)scale;
 - (void) checkNormalsAndAdjustWinding;
 - (void) generateFaceTangents;
 - (void) calculateVertexNormalsAndTangentsWithFaceRefs:(VertexFaceRef *)faceRefs;
@@ -166,13 +164,13 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)object;
 
 // Manage set of objects we need to hang on to, particularly NSDatas owning buffers.
 - (void) setRetainedObject:(id)object forKey:(NSString *)key;
-- (void *) allocateBytesWithSize:(size_t)size count:(OOUInteger)count key:(NSString *)key;
+- (void *) allocateBytesWithSize:(size_t)size count:(NSUInteger)count key:(NSString *)key;
 
 // Allocate all per-vertex/per-face buffers.
-- (BOOL) allocateVertexBuffersWithCount:(OOUInteger)count;
-- (BOOL) allocateNormalBuffersWithCount:(OOUInteger)count;
-- (BOOL) allocateFaceBuffersWithCount:(OOUInteger)count;
-- (BOOL) allocateVertexArrayBuffersWithCount:(OOUInteger)count;
+- (BOOL) allocateVertexBuffersWithCount:(NSUInteger)count;
+- (BOOL) allocateNormalBuffersWithCount:(NSUInteger)count;
+- (BOOL) allocateFaceBuffersWithCount:(NSUInteger)count;
+- (BOOL) allocateVertexArrayBuffersWithCount:(NSUInteger)count;
 
 - (void) renameTexturesFrom:(NSString *)from to:(NSString *)to;
 
@@ -232,13 +230,13 @@ static BOOL IsPerVertexNormalMode(OOMeshNormalMode mode)
 
 @implementation OOMesh
 
-+ (id)meshWithName:(NSString *)name
-		  cacheKey:(NSString *)cacheKey
-materialDictionary:(NSDictionary *)materialDict
- shadersDictionary:(NSDictionary *)shadersDict
-			smooth:(BOOL)smooth
-	  shaderMacros:(NSDictionary *)macros
-shaderBindingTarget:(id<OOWeakReferenceSupport>)object
++ (instancetype) meshWithName:(NSString *)name
+					 cacheKey:(NSString *)cacheKey
+		   materialDictionary:(NSDictionary *)materialDict
+			shadersDictionary:(NSDictionary *)shadersDict
+					   smooth:(BOOL)smooth
+				 shaderMacros:(NSDictionary *)macros
+		  shaderBindingTarget:(id<OOWeakReferenceSupport>)object
 {
 	return [[[self alloc] initWithName:name
 							  cacheKey:cacheKey
@@ -246,7 +244,30 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)object
 					 shadersDictionary:shadersDict
 								smooth:smooth
 						  shaderMacros:macros
-				   shaderBindingTarget:object] autorelease];
+				   shaderBindingTarget:object
+						   scaleFactor:1.0f
+						cacheWriteable:YES] autorelease];
+}
+
++ (instancetype) meshWithName:(NSString *)name
+					 cacheKey:(NSString *)cacheKey
+		   materialDictionary:(NSDictionary *)materialDict
+			shadersDictionary:(NSDictionary *)shadersDict
+					   smooth:(BOOL)smooth
+				 shaderMacros:(NSDictionary *)macros
+		  shaderBindingTarget:(id<OOWeakReferenceSupport>)object
+				  scaleFactor:(float)scale
+			   cacheWriteable:(BOOL)cacheWriteable
+{
+	return [[[self alloc] initWithName:name
+							  cacheKey:cacheKey
+					materialDictionary:materialDict
+					 shadersDictionary:shadersDict
+								smooth:smooth
+						  shaderMacros:macros
+				   shaderBindingTarget:object
+						   scaleFactor:scale
+						cacheWriteable:cacheWriteable] autorelease];
 }
 
 
@@ -269,10 +290,16 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)object
 	if (self == nil)  return nil;
 	
 	baseFile = @"No Model";
+	baseFileOctreeCacheRef = @"No Model-0.000";
+	_cacheWriteable = YES;
 #if OO_MULTITEXTURE
 	_textureUnitCount = NSNotFound;
 #endif
-	
+
+	_lastPosition = kZeroVector;
+	_lastRotMatrix = kZeroMatrix; // not identity
+	_lastBoundingBox = kZeroBoundingBox;
+
 	return self;
 }
 
@@ -281,6 +308,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)object
 {
 	unsigned				i;
 	
+	DESTROY(baseFileOctreeCacheRef);
 	DESTROY(baseFile);
 	DESTROY(octree);
 	
@@ -325,7 +353,7 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 
 - (NSString *)descriptionComponents
 {
-	return [NSString stringWithFormat:@"\"%@\", %u vertices, %u faces, radius: %g m normals: %@", [self modelName], [self vertexCount], [self faceCount], [self collisionRadius], NormalModeDescription(_normalMode)];
+	return [NSString stringWithFormat:@"\"%@\", %zu vertices, %zu faces, radius: %g m normals: %@", [self modelName], [self vertexCount], [self faceCount], [self collisionRadius], NormalModeDescription(_normalMode)];
 }
 
 
@@ -358,19 +386,22 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 {
 	OO_ENTER_OPENGL();
 	
-	OOGL(glPushAttrib(GL_ENABLE_BIT));
+	BOOL meshBelongsToVisualEffect = [_shaderBindingTarget isVisualEffect];
 	
-	OOGL(glShadeModel(GL_SMOOTH));
-	
-	OOGL(glDisableClientState(GL_COLOR_ARRAY));
-	OOGL(glDisableClientState(GL_EDGE_FLAG_ARRAY));
-	
-	OOGL(glEnableClientState(GL_VERTEX_ARRAY));
-	OOGL(glEnableClientState(GL_NORMAL_ARRAY));
+	OOSetOpenGLState(OPENGL_STATE_OPAQUE);
 	
 	OOGL(glVertexPointer(3, GL_FLOAT, 0, _displayLists.vertexArray));
 	OOGL(glNormalPointer(GL_FLOAT, 0, _displayLists.normalArray));
 	
+	// for visual effects enable blending. This will allow use of alpha
+	// channel in shaders - note, this is a bit of cheating the system,
+	// which expects blending to be disabled at this point
+	if (meshBelongsToVisualEffect)
+	{
+		OOGL(glEnable(GL_BLEND));
+		OOGL(glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA));
+	}
+	
 #if OO_SHADERS
 	if ([[OOOpenGLExtensionManager sharedManager] shadersSupported])
 	{
@@ -379,8 +410,6 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 	}
 #endif
 	
-	OOGL(glDisable(GL_BLEND));
-	
 	BOOL usingNormalsAsTexCoords = NO;
 	OOMeshMaterialIndex ti;
 	
@@ -397,12 +426,12 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 		_textureUnitCount = 0;
 		for (ti = 0; ti < materialCount; ti++)
 		{
-			OOUInteger count = [materials[ti] countOfTextureUnitsWithBaseCoordinates];
+			NSUInteger count = [materials[ti] countOfTextureUnitsWithBaseCoordinates];
 			if (_textureUnitCount < count)  _textureUnitCount = count;
 		}
 	}
 	
-	OOUInteger unit;
+	NSUInteger unit;
 	if (_textureUnitCount <= 1)
 	{
 		OOGL(glEnableClientState(GL_TEXTURE_COORD_ARRAY));
@@ -412,7 +441,7 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 		/*	It should not be possible to have multiple texture units if
 			texture combiners are not available.
 		*/
-		NSAssert2([[OOOpenGLExtensionManager sharedManager] textureCombinersSupported], @"Mesh %@ uses %u texture units, but multitexturing is not available.", [self shortDescription], _textureUnitCount);
+		NSAssert2([[OOOpenGLExtensionManager sharedManager] textureCombinersSupported], @"Mesh %@ uses %lu texture units, but multitexturing is not available.", [self shortDescription], _textureUnitCount);
 		
 		for (unit = 0; unit < _textureUnitCount; unit++)
 		{
@@ -424,7 +453,8 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 	OOGL(glEnableClientState(GL_TEXTURE_COORD_ARRAY));
 #endif
 	
-	NS_DURING
+	@try
+	{
 		if (!listsReady)
 		{
 			OOGL(displayList0 = glGenLists(materialCount));
@@ -455,6 +485,11 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 					{
 						OOGL(glDisable(GL_TEXTURE_CUBE_MAP));
 						OOGL(glTexCoordPointer(2, GL_FLOAT, 0, _displayLists.textureUVArray));
+						/*	FIXME: Not including the line below breaks multitexturing in no-shaders mode.
+							However, the OpenGL state manager should probably be handling this;
+							TEXTURE_2D is part of OPENGL_STATE_OPAQUE, which has already been set.
+							- Nikos 20130103
+						*/
 						OOGL(glEnable(GL_TEXTURE_2D));
 					}
 					else
@@ -475,19 +510,17 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 		
 		listsReady = YES;
 		brokenInRender = NO;
-	NS_HANDLER
+	}
+	@catch (NSException *exception)
+	{
 		if (!brokenInRender)
 		{
-			OOLog(kOOLogException, @"***** %s for %@ encountered exception: %@ : %@ *****", __PRETTY_FUNCTION__, self, [localException name], [localException reason]);
+			OOLog(kOOLogException, @"***** %s for %@ encountered exception: %@ : %@ *****", __PRETTY_FUNCTION__, self, [exception name], [exception reason]);
 			brokenInRender = YES;
 		}
-		if ([[localException name] hasPrefix:@"Oolite"])  [UNIVERSE handleOoliteException:localException];	// handle these ourself
-		else  [localException raise];	// pass these on
-	NS_ENDHANDLER
-	
-	OOGL(glDisableClientState(GL_VERTEX_ARRAY));
-	OOGL(glDisableClientState(GL_NORMAL_ARRAY));
-	OOGL(glDisableClientState(GL_TEXTURE_COORD_ARRAY));
+		if ([[exception name] hasPrefix:@"Oolite"])  [UNIVERSE handleOoliteException:exception];	// handle these ourself
+		else  @throw exception;	// pass these on
+	}
 	
 #if OO_SHADERS
 	if ([[OOOpenGLExtensionManager sharedManager] shadersSupported])
@@ -497,12 +530,7 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 #endif
 	
 	[OOMaterial applyNone];
-	CheckOpenGLErrors(@"OOMesh after drawing %@", self);
-	
-#ifndef NDEBUG
-	if (gDebugFlags & DEBUG_DRAW_NORMALS)  [self debugDrawNormals];
-	if (gDebugFlags & DEBUG_OCTREE_DRAW)  [[self octree] drawOctree];
-#endif
+	OOCheckOpenGLErrors(@"OOMesh after drawing %@", self);
 	
 #if OO_MULTITEXTURE
 	if (_textureUnitCount <= 1)
@@ -524,7 +552,15 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 	OOGL(glDisableClientState(GL_TEXTURE_COORD_ARRAY));
 #endif
 	
-	OOGL(glPopAttrib());
+#ifndef NDEBUG
+	if (gDebugFlags & DEBUG_DRAW_NORMALS)  [self debugDrawNormals];
+	if (gDebugFlags & DEBUG_OCTREE_DRAW)  [[self octree] drawOctree];
+#endif
+	
+	// visual effect - disable previously enabled blending
+	if (meshBelongsToVisualEffect)  OOGL(glDisable(GL_BLEND));
+	
+	OOVerifyOpenGLState();
 }
 
 
@@ -601,22 +637,6 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 }
 
 
-- (Geometry *)geometry
-{
-	Geometry *result = [[Geometry alloc] initWithCapacity:faceCount];
-	OOMeshFaceCount i;
-	for (i = 0; i < faceCount; i++)
-	{
-		Triangle tri;
-		tri.v[0] = _vertices[_faces[i].vertex[0]];
-		tri.v[1] = _vertices[_faces[i].vertex[1]];
-		tri.v[2] = _vertices[_faces[i].vertex[2]];
-		[result addTriangle:tri];
-	}
-	return [result autorelease];
-}
-
-
 #if ADAPTIVE_OCTREE_DEPTH
 - (unsigned) octreeDepth
 {
@@ -630,8 +650,7 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 	if (ys < zs)  { t = zs; zs = ys; ys = t; }
 	size = (xs + ys) / 2.0f;	// Use average of two largest
 	
-	if (size < kOctreeVerySmallSizeThreshold)  result = kVerySmallOctreeDepth;
-	else if (size < kOctreeSmallSizeThreshold)  result = kSmallOctreeDepth;
+	if (size < kOctreeSmallSizeThreshold)  result = kSmallOctreeDepth;
 	else while (result < kMaxOctreeDepth)
 	{
 		if (size < threshold) break;
@@ -654,13 +673,36 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 {
 	if (octree == nil)
 	{
-		octree = [OOCacheManager octreeForModel:baseFile];
+		octree = [[OOCacheManager octreeForModel:baseFileOctreeCacheRef] retain];
 		if (octree == nil)
 		{
-			octree = [[self geometry] findOctreeToDepth:[self octreeDepth]];
-			[OOCacheManager setOctree:octree forModel:baseFile];
+			NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
+			
+			OOMeshToOctreeConverter *converter = [OOMeshToOctreeConverter converterWithCapacity:faceCount];
+			OOMeshFaceCount i;
+			for (i = 0; i < faceCount; i++)
+			{
+				// Somewhat surprisingly, this method doesn't even show up in profiles. -- Ahruman 2012-09-22
+				Triangle tri;
+				tri.v[0] = _vertices[_faces[i].vertex[0]];
+				tri.v[1] = _vertices[_faces[i].vertex[1]];
+				tri.v[2] = _vertices[_faces[i].vertex[2]];
+				[converter addTriangle:tri];
+			}
+			
+			octree = [converter findOctreeToDepth:[self octreeDepth]];
+			[octree retain];
+			if (EXPECT(_cacheWriteable))
+			{
+				[OOCacheManager setOctree:octree forModel:baseFileOctreeCacheRef];
+			}
+			
+			[pool release];
+		}
+		else
+		{
+			OOLog(@"mesh.load.octreeCached", @"Retrieved octree \"%@\" from cache.", baseFileOctreeCacheRef);
 		}
-		[octree retain];
 	}
 	
 	return octree;
@@ -717,7 +759,14 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 {
 	// HACK! Should work out what the various bounding box things do and make it neat and consistent.
 	// FIXME: this is a bottleneck.
-	
+// Try to fix bottleneck by caching for common case where subentity
+// pos+rot is constant from frame to frame. - CIM
+
+	if (vector_equal(position,_lastPosition) && OOMatrixEqual(rotMatrix,_lastRotMatrix))
+	{
+		return _lastBoundingBox;
+	}
+
 	BoundingBox		result;
 	Vector			v;
 	
@@ -731,6 +780,10 @@ static NSString *NormalModeDescription(OOMeshNormalMode mode)
 		bounding_box_add_vector(&result,v);
 	}
 	
+	_lastBoundingBox = result;
+	_lastPosition = position;
+	_lastRotMatrix = rotMatrix;
+
 	return result;
 }
 
@@ -824,6 +877,8 @@ materialDictionary:(NSDictionary *)materialDict
 			smooth:(BOOL)smooth
 	  shaderMacros:(NSDictionary *)macros
 shaderBindingTarget:(id<OOWeakReferenceSupport>)target
+	   scaleFactor:(float)scale
+	cacheWriteable:(BOOL)cacheWriteable
 {
 	OOJS_PROFILE_ENTER
 	
@@ -832,17 +887,19 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 	
 	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];
  	_normalMode = smooth ? kNormalModeSmooth : kNormalModePerFace;
+	_cacheWriteable = cacheWriteable;
 	
 #if OOMESH_PROFILE
 	_stopwatch = [[OOProfilingStopwatch alloc] init];
 #endif
 	
-	if ([self loadData:name])
+	if ([self loadData:name scaleFactor:scale])
 	{
 		[self calculateBoundingVolumes];
 		PROFILE(@"finished calculateBoundingVolumes (again\?\?)");
 		
 		baseFile = [name copy];
+		baseFileOctreeCacheRef = [[NSString stringWithFormat:@"%@-%.3f", baseFile, scale] copy];
 		
 		/*	New in r3033: save the material-defining parameters here so we
 			can rebind the materials at any time.
@@ -868,7 +925,10 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 	DESTROY(_stopwatch);
 #endif
 #if OO_MULTITEXTURE
-	_textureUnitCount = NSNotFound;
+	if (EXPECT(self != nil))
+	{
+		_textureUnitCount = NSNotFound;
+	}
 #endif
 	
 	[pool release];
@@ -888,6 +948,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 	if (result != nil)
 	{
 		[result->baseFile retain];
+		[result->baseFileOctreeCacheRef retain];
 		[result->octree retain];
 		[result->_retainedObjects retain];
 		[result->_materialDict retain];
@@ -1098,7 +1159,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 }
 
 
-- (BOOL)loadData:(NSString *)filename
+- (BOOL)loadData:(NSString *)filename scaleFactor:(float)scale
 {
 	OOJS_PROFILE_ENTER
 	
@@ -1111,7 +1172,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 	NSString			*cacheKey = nil;
 	BOOL				using_preloaded = NO;
 	
-	cacheKey = [NSString stringWithFormat:@"%@:%u", filename, _normalMode];
+	cacheKey = [NSString stringWithFormat:@"%@:%u:%.3f", filename, _normalMode, scale];
 	cacheData = [OOCacheManager meshDataForName:cacheKey];
 	if (cacheData != nil)
 	{
@@ -1125,11 +1186,11 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 	
 	if (!using_preloaded)
 	{
-		OOLog(@"mesh.load.uncached", @"Mesh \"%@\" is not in cache, loading.", filename);
+		OOLog(@"mesh.load.uncached", @"Mesh \"%@\" is not in cache, loading.", cacheKey);
 		
 		NSCharacterSet	*whitespaceCharSet = [NSCharacterSet whitespaceCharacterSet];
 		NSCharacterSet	*whitespaceAndNewlineCharSet = [NSCharacterSet whitespaceAndNewlineCharacterSet];
-#if OOLITE_LEOPARD
+#if OOLITE_MAC_OS_X
 		NSCharacterSet	*newlineCharSet = [NSCharacterSet newlineCharacterSet];
 #else
 		static NSCharacterSet *newlineCharSet = nil;
@@ -1150,6 +1211,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 			{
 				// Model not found
 				OOLog(kOOLogMeshDataNotFound, @"***** ERROR: could not find %@", filename);
+				OOStandardsError(@"Model file not found");
 				return NO;
 			}
 			
@@ -1263,7 +1325,7 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 					if (![scanner scanFloat:&z])  failFlag = YES;
 					if (!failFlag)
 					{
-						_vertices[j] = make_vector(x, y, z);
+						_vertices[j] = make_vector(x*scale, y*scale, z*scale);
 					}
 					else
 					{
@@ -1320,11 +1382,11 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 						if (n_v < 3)
 						{
 							failFlag = YES;
-							failString = [NSString stringWithFormat:@"%@Face[%u] has fewer than three vertices.\n", failString];
+							failString = [NSString stringWithFormat:@"%@Face[%u] has fewer than three vertices.\n", failString, j];
 						}
 						else if (n_v > 3)
 						{
-							OOLogWARN(@"mesh.load.warning.nonTriangular", @"Face[%u] of %@ has %u vertices specified. Only the first three will be used.", baseFile, n_v);
+							OOLogWARN(@"mesh.load.warning.nonTriangular", @"Face[%u] of %@ has %u vertices specified. Only the first three will be used.", j, baseFile, n_v);
 							n_v = 3;
 						}
 					}
@@ -1557,8 +1619,11 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 		}
 		
 		// save the resulting data for possible reuse
-		[OOCacheManager setMeshData:[self modelData] forName:cacheKey];
-		PROFILE(@"saved to cache");
+		if (EXPECT(_cacheWriteable))
+		{
+			[OOCacheManager setMeshData:[self modelData] forName:cacheKey];
+			PROFILE(@"saved to cache");
+		}
 		
 		if (failFlag)
 		{
@@ -1659,33 +1724,25 @@ shaderBindingTarget:(id<OOWeakReferenceSupport>)target
 		
 		if (dsAC * dtAB > dsAB * dtAC)
 		{
-			dsAB = -dsAB;
-			dsAC = -dsAC;
+			face->tangent = vector_normal(vector_subtract(vector_multiply_scalar(vProjAC, dtAB), vector_multiply_scalar(vProjAB, dtAC)));
 		}
-		
-		Vector tangent = vector_subtract(vector_multiply_scalar(vProjAB, dsAC), vector_multiply_scalar(vProjAC, dsAB));
-		face->tangent = cross_product(nA, tangent);	// Rotate 90 degrees. Done this way because I'm too lazy to grok the code above.
+		else
+		{
+			face->tangent = vector_normal(vector_subtract(vector_multiply_scalar(vProjAB, dtAC), vector_multiply_scalar(vProjAC, dtAB)));
+		}			
 	}
 	
 	OOJS_PROFILE_EXIT_VOID
 }
 
 
-static float FaceAreaBroken(GLuint *vertIndices, Vector *vertices)
+static float FaceArea(GLuint *vertIndices, Vector *vertices)
 {
-	/*	This is supposed to calculate areas using Heron's formula, but doesn't.
-		(The *0.25 is supposed to be outside the sqrt.) This bug was introduced
-		somewhere between version 1.40 and 1.55, so we can't really fix it at
-		this point.
-		Current plan: replace DAT files with a better format post-MNSR. The
-		format converter will be responsible for any smoothing and can do it
-		right.
-		-- Ahruman 2010-05-22
-	*/
+	/*	Calculate areas using Heron's formula.	*/
 	float	a2 = distance2(vertices[vertIndices[0]], vertices[vertIndices[1]]);
 	float	b2 = distance2(vertices[vertIndices[1]], vertices[vertIndices[2]]);
 	float	c2 = distance2(vertices[vertIndices[2]], vertices[vertIndices[0]]);
-	return sqrtf(2.0 * (a2 * b2 + b2 * c2 + c2 * a2) - 0.25 * (a2 * a2 + b2 * b2 +c2 * c2));
+	return sqrt((2.0f * (a2 * b2 + b2 * c2 + c2 * a2) - (a2 * a2 + b2 * b2 +c2 * c2)) * 0.0625f);
 }
 
 
@@ -1695,14 +1752,14 @@ static float FaceAreaBroken(GLuint *vertIndices, Vector *vertices)
 	
 	NSParameterAssert(faceRefs != NULL);
 	
-	OOUInteger	i,j;
+	NSUInteger	i,j;
 	float		triangle_area[faceCount];
 	
 	NSAssert1(_normals != NULL && _tangents != NULL, @"Normal/tangent buffers not allocated in %s", __PRETTY_FUNCTION__);
 	
 	for (i = 0 ; i < faceCount; i++)
 	{
-		triangle_area[i] = FaceAreaBroken(_faces[i].vertex, _vertices);
+		triangle_area[i] = FaceArea(_faces[i].vertex, _vertices);
 	}
 	for (i = 0; i < vertexCount; i++)
 	{
@@ -1710,7 +1767,7 @@ static float FaceAreaBroken(GLuint *vertIndices, Vector *vertices)
 		Vector tangent_sum = kZeroVector;
 		
 		VertexFaceRef *vfr = &faceRefs[i];
-		OOUInteger fIter, fCount = VFRGetCount(vfr);
+		NSUInteger fIter, fCount = VFRGetCount(vfr);
 		for (fIter = 0; fIter < fCount; fIter++)
 		{
 			j = VFRGetFaceAtIndex(vfr, fIter);
@@ -1721,6 +1778,7 @@ static float FaceAreaBroken(GLuint *vertIndices, Vector *vertices)
 		}
 		
 		normal_sum = vector_normal_or_fallback(normal_sum, kBasisZVector);
+		tangent_sum = vector_subtract(tangent_sum, vector_multiply_scalar(normal_sum, dot_product(tangent_sum, normal_sum)));
 		tangent_sum = vector_normal_or_fallback(tangent_sum, kBasisXVector);
 		
 		_normals[i] = normal_sum;
@@ -1766,7 +1824,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 		smooth-grouped normal generation.
 		-- Ahruman 2010-05-22
 	*/
-	OOUInteger	i,j;
+	NSUInteger	i,j;
 	float	triangle_area[faceCount];
 	for (i = 0 ; i < faceCount; i++)
 	{
@@ -1777,7 +1835,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 		Vector tangent_sum = kZeroVector;
 		
 		VertexFaceRef *vfr = &faceRefs[i];
-		OOUInteger fIter, fCount = VFRGetCount(vfr);
+		NSUInteger fIter, fCount = VFRGetCount(vfr);
 		for (fIter = 0; fIter < fCount; fIter++)
 		{
 			j = VFRGetFaceAtIndex(vfr, fIter);
@@ -1786,6 +1844,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 			tangent_sum = vector_add(tangent_sum, vector_multiply_scalar(_faces[j].tangent, t));
 		}
 		
+		tangent_sum = vector_subtract(tangent_sum, vector_multiply_scalar(_normals[i], dot_product(_normals[i], tangent_sum)));
 		tangent_sum = vector_normal_or_fallback(tangent_sum, kBasisXVector);
 		
 		_tangents[i] = tangent_sum;
@@ -1794,14 +1853,15 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 	OOJS_PROFILE_EXIT_VOID
 }
 
-
+/* profiling suggests this function takes a lot of time - almost all
+ * the overhead of setting up a new ship is here. - CIM */
 - (void) getNormal:(Vector *)outNormal andTangent:(Vector *)outTangent forVertex:(OOMeshVertexCount)v_index inSmoothGroup:(OOMeshSmoothGroup)smoothGroup
 {
 	OOJS_PROFILE_ENTER
 	
 	assert(outNormal != NULL && outTangent != NULL);
 	
-	OOUInteger j;
+	NSUInteger j;
 	Vector normal_sum = kZeroVector;
 	Vector tangent_sum = kZeroVector;
 	for (j = 0; j < faceCount; j++)
@@ -1810,7 +1870,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 		{
 			if ((_faces[j].vertex[0] == v_index)||(_faces[j].vertex[1] == v_index)||(_faces[j].vertex[2] == v_index))
 			{
-				float area = FaceAreaBroken(_faces[j].vertex, _vertices);
+				float area = FaceArea(_faces[j].vertex, _vertices);
 				normal_sum = vector_add(normal_sum, vector_multiply_scalar(_faces[j].normal, area));
 				tangent_sum = vector_add(tangent_sum, vector_multiply_scalar(_faces[j].tangent, area));
 			}
@@ -1828,7 +1888,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 {
 	OOJS_PROFILE_ENTER
 	
-	OOUInteger	fi, vi, mi;
+	NSUInteger	fi, vi, mi;
 	
 	if (![self allocateVertexArrayBuffersWithCount:faceCount])  return NO;
 	
@@ -1924,7 +1984,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 	OOJS_PROFILE_ENTER
 	
 	OOMeshVertexCount	i;
-	double				d_squared, length_longest_axis, length_shortest_axis;
+	float				d_squared, length_longest_axis, length_shortest_axis;
 	GLfloat				result;
 	
 	result = 0.0f;
@@ -1953,7 +2013,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 	d_squared = (length_longest_axis + length_shortest_axis) * (length_longest_axis + length_shortest_axis) * 0.25; // square of average length
 	maxDrawDistance = d_squared * NO_DRAW_DISTANCE_FACTOR * NO_DRAW_DISTANCE_FACTOR;	// no longer based on the collision radius
 	
-	collisionRadius = sqrt(result);
+	collisionRadius = sqrtf(result);
 	
 	OOJS_PROFILE_EXIT_VOID
 }
@@ -1965,14 +2025,14 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 	OOMeshVertexCount	i;
 	Vector				*vertex = NULL;
 	
-	for (i = 0; i != vertexCount; ++i)
+	for (i = 0; i < vertexCount; i++)
 	{
 		vertex = &_vertices[i];
 		*vertex = vector_multiply_scalar(*vertex, factor);
 	}
 	
 	// Rescale actual display vertices.
-	for (i = 0; i != _displayLists.count; i++)
+	for (i = 0; i < _displayLists.count; i++)
 	{
 		vertex = &_displayLists.vertexArray[i];
 		*vertex = vector_multiply_scalar(*vertex, factor);
@@ -1981,6 +2041,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 	[self calculateBoundingVolumes];
 	DESTROY(octree);
 	DESTROY(baseFile);	// Avoid octree cache.
+	DESTROY(baseFileOctreeCacheRef);
 }
 
 
@@ -2014,7 +2075,7 @@ static float FaceAreaCorrect(GLuint *vertIndices, Vector *vertices)
 		
 		// Draw normal
 		length = magnitude2(n);
-		blend = fabsf(length - 1) * 5.0;
+		blend = fabs(length - 1) * 5.0;
 		color[0] = MIN(blend, 1.0f);
 		color[1] = 1.0f - color[0];
 		color[2] = color[1];
@@ -2074,8 +2135,12 @@ static void Scribble(void *bytes, size_t size)
 #endif
 
 
-- (void *) allocateBytesWithSize:(size_t)size count:(OOUInteger)count key:(NSString *)key
+/* valgrind complains that the memory allocated here isn't initialised
+	 at the time OOCacheManager::writeDict is pushing it to the
+	 cache. Not sure if that's a problem or not. - CIM */
+- (void *) allocateBytesWithSize:(size_t)size count:(NSUInteger)count key:(NSString *)key
 {
+	if (count == 0) { count=1; }
 	size *= count;
 	void *bytes = malloc(size);
 	if (bytes != NULL)
@@ -2088,14 +2153,14 @@ static void Scribble(void *bytes, size_t size)
 }
 
 
-- (BOOL) allocateVertexBuffersWithCount:(OOUInteger)count
+- (BOOL) allocateVertexBuffersWithCount:(NSUInteger)count
 {
 	_vertices = [self allocateBytesWithSize:sizeof *_vertices count:vertexCount key:@"vertices"];
 	return _vertices != NULL;
 }
 
 
-- (BOOL) allocateNormalBuffersWithCount:(OOUInteger)count
+- (BOOL) allocateNormalBuffersWithCount:(NSUInteger)count
 {
 	_normals = [self allocateBytesWithSize:sizeof *_normals count:vertexCount key:@"normals"];
 	_tangents = [self allocateBytesWithSize:sizeof *_tangents count:vertexCount key:@"tangents"];
@@ -2103,14 +2168,14 @@ static void Scribble(void *bytes, size_t size)
 }
 
 
-- (BOOL) allocateFaceBuffersWithCount:(OOUInteger)count
+- (BOOL) allocateFaceBuffersWithCount:(NSUInteger)count
 {
 	_faces = [self allocateBytesWithSize:sizeof *_faces count:faceCount key:@"faces"];
 	return	_faces != NULL;
 }
 
 
-- (BOOL) allocateVertexArrayBuffersWithCount:(OOUInteger)count
+- (BOOL) allocateVertexArrayBuffersWithCount:(NSUInteger)count
 {
 	_displayLists.indexArray = [self allocateBytesWithSize:sizeof *_displayLists.indexArray count:count * 3 key:@"indexArray"];
 	_displayLists.textureUVArray = [self allocateBytesWithSize:sizeof *_displayLists.textureUVArray count:count * 6 key:@"textureUVArray"];
@@ -2198,7 +2263,7 @@ static NSString * const kOOCacheOctrees = @"octrees";
 }
 
 
-static void VFRAddFace(VertexFaceRef *vfr, OOUInteger index)
+static void VFRAddFace(VertexFaceRef *vfr, NSUInteger index)
 {
 	NSCParameterAssert(vfr != NULL);
 	
@@ -2209,12 +2274,12 @@ static void VFRAddFace(VertexFaceRef *vfr, OOUInteger index)
 	else
 	{
 		if (vfr->extra == nil)  vfr->extra = [NSMutableArray array];
-		[vfr->extra addObject:[NSNumber numberWithInt:index]];
+		[vfr->extra addObject:[NSNumber numberWithInteger:index]];
 	}
 }
 
 
-static OOUInteger VFRGetCount(VertexFaceRef *vfr)
+static NSUInteger VFRGetCount(VertexFaceRef *vfr)
 {
 	NSCParameterAssert(vfr != NULL);
 	
@@ -2222,7 +2287,7 @@ static OOUInteger VFRGetCount(VertexFaceRef *vfr)
 }
 
 
-static OOUInteger VFRGetFaceAtIndex(VertexFaceRef *vfr, OOUInteger index)
+static NSUInteger VFRGetFaceAtIndex(VertexFaceRef *vfr, NSUInteger index)
 {
 	NSCParameterAssert(vfr != NULL && index < VFRGetCount(vfr));
 	
