@@ -3,7 +3,7 @@
 OOJavaScriptEngine.h
 
 JavaScript support for Oolite
-Copyright (C) 2007-2011 David Taylor and Jens Ayton.
+Copyright (C) 2007-2013 David Taylor and Jens Ayton.
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
@@ -30,7 +30,7 @@ MA 02110-1301, USA.
 #include <jsapi.h>
 
 
-#define OOJSENGINE_MONITOR_SUPPORT	(!defined(NDEBUG))
+#define OOJSENGINE_MONITOR_SUPPORT OOLITE_DEBUG
 
 
 #import "OOJSPropID.h"
@@ -82,7 +82,7 @@ MA 02110-1301, USA.
 - (void) removeGCObjectRoot:(JSObject **)rootPtr;
 - (void) removeGCValueRoot:(jsval *)rootPtr;
 
-- (void) garbageCollectionOpportunity;
+- (void) garbageCollectionOpportunity:(BOOL)force;
 
 - (BOOL) showErrorLocations;
 - (void) setShowErrorLocations:(BOOL)value;
@@ -323,21 +323,14 @@ jsid OOJSIDFromString(NSString *string);
 @interface NSString (OOJavaScriptExtensions)
 
 // For diagnostic messages; produces things like @"(42, true, "a string", an object description)".
-+ (id) stringWithJavaScriptParameters:(jsval *)params count:(uintN)count inContext:(JSContext *)context;
++ (NSString *) stringWithJavaScriptParameters:(jsval *)params count:(uintN)count inContext:(JSContext *)context;
 
 // Concatenate sequence of arbitrary JS objects into string.
-+ (id) concatenationOfStringsFromJavaScriptValues:(jsval *)values count:(size_t)count separator:(NSString *)separator inContext:(JSContext *)context;
++ (NSString *) concatenationOfStringsFromJavaScriptValues:(jsval *)values count:(size_t)count separator:(NSString *)separator inContext:(JSContext *)context;
 
 // Add escape codes for string so that it's a valid JavaScript literal (if you put "" or '' around it).
 - (NSString *) escapedForJavaScriptLiteral;
 
-
-// Wrapper for OOStringFromJSValueEvenIfNull(). DEPRECATED
-+ (id) stringWithJavaScriptValue:(jsval)value inContext:(JSContext *)context;
-
-// Wrapper for OOStringFromJSValue(). DEPRECATED
-+ (id) stringOrNilWithJavaScriptValue:(jsval)value inContext:(JSContext *)context;
-
 @end
 
 
@@ -359,6 +352,9 @@ BOOL JSEntityIsJavaScriptVisiblePredicate(Entity *entity, void *parameter);
 // YES for ships other than sub-entities and menu-display ships, and planets other than atmospheres and menu miniatures. Parameter: ignored.
 BOOL JSEntityIsJavaScriptSearchablePredicate(Entity *entity, void *parameter);
 
+// YES for menu-display ships. Parameter: ignored
+BOOL JSEntityIsDemoShipPredicate(Entity *entity, void *parameter);
+
 
 // These require a request on context.
 id OOJSNativeObjectFromJSValue(JSContext *context, jsval value);
@@ -445,7 +441,7 @@ NSDictionary *OOJSDictionaryFromStringTable(JSContext *context, jsval value);
 	The signature of the generator is:
 	BOOL <name>(JSContext *context, JSObject *inObject, <class>** outObject)
 	If it returns NO, inObject is of the wrong class and an error has been
-	raised. Otherwise, outOjbect is either a native object of the specified
+	raised. Otherwise, outObject is either a native object of the specified
 	class (or a subclass) or nil.
 */
 #ifndef NDEBUG
@@ -558,7 +554,7 @@ void OOJSRegisterObjectConverter(JSClass *theClass, OOJSClassConverterCallback c
 		showingLocation:(in BOOL)showLocation
 			withMessage:(in NSString *)message;
 
-// Sent for JS log messages. Note: messageClass will be nil of Log() is used rather than LogWithClass().
+// Sent for JS log messages. Note: messageClass will be nil if Log() is used rather than LogWithClass().
 - (oneway void)jsEngine:(in byref OOJavaScriptEngine *)engine
 				context:(in JSContext *)context
 			 logMessage:(in NSString *)message
@@ -682,5 +678,6 @@ do { \
 } while (0)
 
 #define OOJS_RETURN_VECTOR(value)		OOJS_RETURN_WITH_HELPER(VectorToJSValue, value)
+#define OOJS_RETURN_HPVECTOR(value)		OOJS_RETURN_WITH_HELPER(HPVectorToJSValue, value)
 #define OOJS_RETURN_QUATERNION(value)	OOJS_RETURN_WITH_HELPER(QuaternionToJSValue, value)
 #define OOJS_RETURN_DOUBLE(value)		OOJS_RETURN_WITH_HELPER(JS_NewNumberValue, value)
